"""wrapper methods to work with S3"""
import boto3
import gzip
import os
import zipfile
from io import BytesIO

# local import
from omigo_core import utils
import threading
import traceback

# define some global variables for caching s3 bucket and session
S3_RESOURCE = {}
S3_SESSIONS = {}
S3_BUCKETS = {}
S3_CLIENTS = {}
S3_RESOURCE_LOCK = threading.Lock()
S3_SESSION_LOCK = threading.Lock()
S3_CLIENT_LOCK = threading.Lock()

def create_session_key(s3_region = None, aws_profile = None):
    s3_region, aws_profile = resolve_region_profile(s3_region, aws_profile)
    if (s3_region is None and aws_profile is None):
        return "DEFAULT_KEY"
    else:
        s3_region_str = "" if (s3_region is None) else s3_region
        aws_profile_str = "" if (aws_profile is None) else aws_profile
        return "{}:{}".format(s3_region_str, aws_profile_str)

def get_s3_session(s3_region = None, aws_profile = None):
    s3_region, aws_profile = resolve_region_profile(s3_region, aws_profile)

    # show the s3 settings
    utils.info_once("get_s3_session: s3_region: {}, aws_profile: {}".format(s3_region, aws_profile))

    # generate s3_session
    if (s3_region is not None and aws_profile is not None):
        session = boto3.session.Session(region_name = s3_region, profile_name = aws_profile)
        utils.debug("get_s3_session: s3_region: {}, aws_profile: {}, session: {}".format(s3_region, aws_profile, session))
    else:
        session = boto3.session.Session(region_name = s3_region, profile_name = aws_profile)
        utils.debug("get_s3_session: no s3_region or aws_profile, session: {}".format(session))

    # return
    return session

def get_s3_session_cache(s3_region = None, aws_profile = None):
    s3_region, aws_profile = resolve_region_profile(s3_region, aws_profile)
    key = create_session_key(s3_region, aws_profile)

    # make it thread safe
    with S3_SESSION_LOCK:
        if ((key in S3_SESSIONS.keys()) == False):
            S3_SESSIONS[key] = get_s3_session(s3_region, aws_profile)

    utils.debug("get_s3_session_cache: s3_region: {}, aws_profile: {}, key: {}, sessions: {}".format(s3_region, aws_profile, key, S3_SESSIONS))
    return S3_SESSIONS[key]

# TODO
def get_s3_resource(s3_region = None, aws_profile = None):
    s3_region, aws_profile = resolve_region_profile(s3_region, aws_profile)
    session = get_s3_session_cache(s3_region, aws_profile)

    return session.resource("s3")

def get_s3_resource_cache(s3_region = None, aws_profile = None):
    s3_region, aws_profile = resolve_region_profile(s3_region, aws_profile)
    key = create_session_key(s3_region, aws_profile)

    # make it thread safe
    with S3_RESOURCE_LOCK:
        if ((key in S3_RESOURCE.keys()) == False):
            S3_RESOURCE[key] = get_s3_resource(s3_region, aws_profile)

    # return
    return S3_RESOURCE[key]

def get_s3_client(s3_region = None, aws_profile = None):
    s3_region, aws_profile = resolve_region_profile(s3_region, aws_profile)
    session = get_s3_session_cache(s3_region, aws_profile)
    return session.client("s3")

def get_s3_client_cache(s3_region = None, aws_profile = None):
    s3_region, aws_profile = resolve_region_profile(s3_region, aws_profile)
    key = create_session_key(s3_region, aws_profile)

    # make it thread safe
    with S3_CLIENT_LOCK:
        if ((key in S3_CLIENTS.keys()) == False):
            S3_CLIENTS[key] = get_s3_client(s3_region, aws_profile)

    # return
    return S3_CLIENTS[key]

def get_s3_bucket(bucket_name, s3_region = None, aws_profile = None):
    s3_region, aws_profile = resolve_region_profile(s3_region, aws_profile)
    s3 = get_s3_resource_cache(s3_region, aws_profile)

    return s3.Bucket(bucket_name)

def get_s3_bucket_cache(bucket_name, s3_region = None, aws_profile = None):
    s3_region, aws_profile = resolve_region_profile(s3_region, aws_profile)
    if ((bucket_name in S3_BUCKETS.keys()) == False):
        S3_BUCKETS[bucket_name] = get_s3_bucket(bucket_name, s3_region, aws_profile)

    # return
    return S3_BUCKETS[bucket_name]

def get_file_content(bucket_name, object_key, s3_region = None, aws_profile = None):
    s3_region, aws_profile = resolve_region_profile(s3_region, aws_profile)
    s3_bucket = get_s3_bucket_cache(bucket_name, s3_region, aws_profile)
    obj = s3_bucket.Object(object_key)
    obj_content = obj.get()
    body = obj_content["Body"]
    data = body.read()
    body.close()

    return data

# TODO: Deprecated
def get_s3_file_content(bucket_name, object_key, s3_region = None, aws_profile = None):
    utils.warn_once("use get_file_content instead")
    return get_file_content(bucket_name, object_key, s3_region = s3_region, aws_profile = aws_profile)

# TODO: Deprecated
def get_s3_file_content_as_text(bucket_name, object_key, s3_region = None, aws_profile = None):
    utils.warn_once("use get_file_content_as_text instead")
    return get_file_content_as_text(bucket_name, object_key, s3_region = s3_region, aws_profile = aws_profile)

def get_file_content_as_text(bucket_name, object_key, s3_region = None, aws_profile = None):
    s3_region, aws_profile = resolve_region_profile(s3_region, aws_profile)
    barr = get_file_content(bucket_name, object_key, s3_region, aws_profile)

    # check for gz or zip
    if (object_key.endswith(".gz")):
        barr = gzip.decompress(barr)
    elif(object_key.endswith(".zip")):
        # get only the filename
        zfile = zipfile.ZipFile(BytesIO(barr))
        barr = zfile.open(zfile.infolist()[0]).read()
        zfile.close()

    # return bytearray
    barr = bytearray(barr)
    return barr.decode().rstrip("\n")

# TODO: this is expensive and works in specific scenarios only especially for files
def check_path_exists(path, s3_region = None, aws_profile = None):
    s3_region, aws_profile = resolve_region_profile(s3_region, aws_profile)

    s3 = get_s3_client_cache(s3_region = s3_region, aws_profile = aws_profile)
    bucket_name, object_key = utils.split_s3_path(path)

    # show the s3 settings
    utils.info_once("check_path_exists: s3_region: {}, aws_profile: {}".format(s3_region, aws_profile))
    # utils.info_once("check_path_exists: s3_region: {}, aws_profile: {}, s3: {}".format(s3_region, aws_profile, s3))

    results = s3.list_objects(Bucket = bucket_name, Prefix = object_key)
    return "Contents" in results

def check_file_exists(path, s3_region = None, aws_profile = None):
    s3_region, aws_profile = resolve_region_profile(s3_region, aws_profile)
    s3 = get_s3_client_cache(s3_region = s3_region, aws_profile = aws_profile)
    bucket_name, object_key = utils.split_s3_path(path)
    try:
        # call head_object which is supposed to be efficient. This works only for files and not directories?
        response = s3.head_object(Bucket = bucket_name, Key = object_key)

        # check response
        if (response is not None):
            return True
        else:
            return False
    except Exception as e:
        # utils.error("check_file_exists: path: {}, s3_region: {}, aws_profile: {}, exception: {}".format(path, s3_region, aws_profile, e))
        # utils.error("check_file_exists: StackTrace: {}".format(traceback.format_exc()))
        return False

def put_file_content(bucket_name, object_key, barr, s3_region = None, aws_profile = None):
    # get s3 references
    s3_region, aws_profile = resolve_region_profile(s3_region, aws_profile)
    s3_bucket = get_s3_resource_cache(s3_region, aws_profile)

    # get object
    obj = s3_bucket.Object(bucket_name, object_key)

    # write
    obj.put(Body = barr)

# TODO: Deprecated
def put_s3_file_content(bucket_name, object_key, barr, s3_region = None, aws_profile = None):
    utils.warn_once("use put_file_content instead")
    return put_file_content(bucket_name, object_key, barr, s3_region = None, aws_profile = None)

def put_file_with_text_content(bucket_name, object_key, text, s3_region = None, aws_profile = None):
    s3_region, aws_profile = resolve_region_profile(s3_region, aws_profile)
    barr = str.encode(text)
    if (object_key.endswith(".gz")):
        barr = gzip.compress(barr)
    elif (object_key.endswith(".zip")):
        # get only the filename
        object_file_name = object_key.split("/")[-1]
        mzip = BytesIO()
        with zipfile.ZipFile(mzip, mode = "w", compression = zipfile.ZIP_DEFLATED) as zfile:
            zfile.writestr(object_file_name[0:-4], barr)
        barr = mzip.getvalue()

    put_file_content(bucket_name, object_key, barr, s3_region = s3_region, aws_profile = aws_profile)

# TODO: Deprecated
def put_s3_file_with_text_content(bucket_name, object_key, text, s3_region = None, aws_profile = None):
    utils.warn_once("use put_file_with_text_content instead")
    put_file_with_text_content(bucket_name, object_key, text, s3_region = s3_region, aws_profile = aws_profile)

def resolve_region_profile(s3_region = None, aws_profile = None):
    # resolve s3_region
    if ((s3_region is None or s3_region == "") and "S3_REGION" in os.environ.keys()):
        s3_region = os.environ["S3_REGION"]

    # resolve aws_profile
    if ((aws_profile is None or aws_profile == "") and "AWS_PROFILE" in os.environ.keys()):
        aws_profile = os.environ["AWS_PROFILE"]

    # return
    return s3_region, aws_profile

# https://stackoverflow.com/questions/54314563/how-to-get-more-than-1000-objects-from-s3-by-using-list-objects-v2
def __get_all_s3_objects__(s3, **base_kwargs):
    continuation_token = None
    while True:
        list_kwargs = dict(MaxKeys=1000, **base_kwargs)
        if continuation_token:
            list_kwargs['ContinuationToken'] = continuation_token
        response = s3.list_objects_v2(**list_kwargs)
        yield from response.get('Contents', [])
        if not response.get('IsTruncated'):  # At the end of the list?
            break
        continuation_token = response.get('NextContinuationToken')

# FIXME: This method is implemented using reverse engineering. Not so reliable
# TODO: ignore_if_missing should be FALSE by default
# FIXME: the prefix in aws s3 list command are hurting
def get_directory_listing(path, filter_func = None, ignore_if_missing = False, skip_exist_check = False, s3_region = None, aws_profile = None):
    s3_region, aws_profile = resolve_region_profile(s3_region, aws_profile)
    s3 = get_s3_client_cache(s3_region = s3_region, aws_profile = aws_profile)

    # split the path
    bucket_name, object_key = utils.split_s3_path(path)
    # utils.info("get_directory_listing: bucket_name: {}, object_key: {}".format(bucket_name, object_key))

    filenames = []
    # validation
    if (check_path_exists(path, s3_region = s3_region, aws_profile = aws_profile) == False):
        if (ignore_if_missing == False):
            raise Exception("Directory does not exist: " + path)
        else:
            utils.debug("Directory does not exist: {}".format(path))
            return []

    # set the max keys as boto3 api are weird
    response = __get_all_s3_objects__(s3, Bucket = bucket_name, Prefix = object_key)

    # iterate through response
    count = 0
    for dir_content in response:
        count = count + 1
        # utils.debug("Response: {}".format(dir_content))
        filename = dir_content['Key']

        # check if the directory listing was done on leaf node in which case ignore
        if (object_key == filename):
            continue

        # check if there was a prefix match to a parallel directory or key at the same level. in that case ignore
        if (filename.startswith(object_key)):
            filename_remaining = filename[len(object_key):]

            # check for valid prefix to a parallel directory
            if (len(filename_remaining) > 0 and filename_remaining.startswith("/") == False):
                continue

        # extract the key for remaining
        extracted_key = filename[len(object_key) + 1:]
        # utils.debug("extracted_key: {}".format(extracted_key))

        # this had directory as child
        parts = extracted_key.split("/")
        base_filename = None
        if (len(parts) == 1):
            # immediate child is a leaf node
            base_filename = parts[0]
        else:
            # immediate child is a directory
            base_filename = parts[0]

        # debug
        # utils.debug("object_key: {}, filename: {}, extracted_key: {}, base_filename: {}".format(object_key, filename, extracted_key, base_filename))

        # collect all the paths found
        if (base_filename is not None):
            # extract the last part and then prepend the bucket and object key to construct full path
            filenames.append(path + "/" + base_filename)

    # print the number of object_keys found
    # utils.trace("Number of entries found in directory listing: {}: {}".format(path, count))

    # dedup
    filenames = sorted(list(set(filenames)))

    # valid files as boto does a prefix search
    if (skip_exist_check == False):
        filenames = list(filter(lambda t: check_path_exists(t, s3_region = s3_region, aws_profile = aws_profile), filenames))

    # apply filter func if any
    if (filter_func is not None):
        filenames = list(filter(lambda t: filter_func(t), filenames))

    # return
    return filenames

def delete_file(path, ignore_if_missing = False, s3_region = None, aws_profile = None):
    s3_region, aws_profile = resolve_region_profile(s3_region, aws_profile)
    s3 = get_s3_client_cache(s3_region = s3_region, aws_profile = aws_profile)

    # split the path
    bucket_name, object_key = utils.split_s3_path(path)

    # check if the file exists
    if (check_path_exists(path, s3_region = s3_region, aws_profile = aws_profile) == False):
        if (ignore_if_missing == False):
            raise Exception("delete_file: path doesnt exist: {}".format(path))
        else:
            utils.debug("delete_file: path doesnt exist: {}".format(path))

        return

    # delete
    s3.delete_object(Bucket = bucket_name, Key = object_key)

def get_last_modified_time(path, ignore_if_missing = False, s3_region = None, aws_profile = None):
    # check if exists
    if (check_path_exists(path, s3_region = s3_region, aws_profile = aws_profile) == False):
        if (ignore_if_missing == False):
            raise Exception("get_last_modified_time: path doesnt exist: {}".format(path))
        else:
            utils.debug("get_last_modified_time: path doesnt exist: {}".format(path))

        return None

    # parse
    s3_region, aws_profile = resolve_region_profile(s3_region, aws_profile)
    s3 = get_s3_client_cache(s3_region = s3_region, aws_profile = aws_profile)

    # split the path
    bucket_name, object_key = utils.split_s3_path(path)

    # call head_object to get metadata
    response = s3.head_object(Bucket = bucket_name, Key = object_key)

    # validation
    if (response is not None):
        # get datetime
        datetime_value = response["LastModified"]

        # return
        return datetime_value
    else:
        return None



import os
import datetime
from omigo_core import timefuncs
from omigo_core import utils
import zipfile, gzip

def check_path_exists(path):
    return os.path.exists(path)

def check_file_exists(path):
    return os.path.exists(path)

def get_directory_listing(path, filter_func = None, ignore_if_missing = False, skip_exist_check = False):
    # skip_exist_check is for performance and parity with s3
    if (skip_exist_check == False):
        if (check_path_exists(path) == False):
            if (ignore_if_missing == False):
                raise Exception("Directory does not exist: {}".format(path))
            else:
                return []

    full_paths = []
    for p in os.listdir(path):
        full_paths.append(path + "/" + p)

    # apply filter func if any
    if (filter_func is not None):
        full_paths = list(filter(lambda t: filter_func(t), full_paths))

    # return
    return full_paths

# TODO: not in s3
def makedirs(path, exist_ok = True):
    return os.makedirs(path, exist_ok = exist_ok)

def get_file_content_as_text(path):
    # initialize
    data = None

    # check for file exists
    if (check_path_exists(path) == False):
        raise Exception("file doesnt exist: {}".format(path))

    # read based on file type
    if (path.endswith(".gz")):
        fin = gzip.open(path, mode = "rt")
        data = [x.rstrip("\n") for x in fin.readlines()]
        fin.close()
    elif (path.endswith(".zip")):
        zipf = zipfile.ZipFile(path, "r")
        fin = zipf.open(zipf.infolist()[0], "r")
        data = fin.read().decode().split("\n")
        fin.close()
        zipf.close()
    else:
        fin = open(path, "r")
        data = [x.rstrip("\n") for x in fin.readlines()]
        fin.close()

    # return
    return "\n".join(data)

def get_last_modified_timestamp(path):
    return timefuncs.datetime_to_utctimestamp_sec(datetime.datetime.utcfromtimestamp(int(os.path.getmtime(path))))

# TODO: this is not tested
def put_file_with_text_content(path, text):
    utils.warn_once("put_file_with_text_content is not tested fully")

    # initialize
    output_zipf = None
    output_file = None

    # construct output file
    if (path.endswith(".gz")):
        output_file = gzip.open(path, "wt")
    elif (path.endswith(".zip")):
        output_zipf = zipfile.ZipFile(path, "wt")
        output_file = output_zipf.open(path.split("/")[-1][0:-4], "w")
    else:
        output_file = open(path, "wt")

    # write
    output_file.write(text)

    # close
    if (output_zipf is not None):
        output_zipf.close()
    if (output_file is not None):
        output_file.close()

# TODO: return success status
def delete_file(path, ignore_if_missing = True):
    utils.debug("delete_file: path: {}, ignore_if_missing: {}".format(path, ignore_if_missing))
    # check if the file exists
    if (check_path_exists(path) == False):
        if (ignore_if_missing == False):
            raise Exception("delete_file: path doesnt exist: {}".format(path))
        else:
            utils.debug("delete_file: path doesnt exist: {}".format(path))

        return

    # delete
    os.remove(path)

# TODO: This api doesnt have s3 counterpart
def delete_dir(path, ignore_if_missing = True):
    utils.warn_once("delete_dir: this api doesnt have s3 counterpart")
    utils.debug("delete_dir: path: {}, ignore_if_missing: {}".format(path, ignore_if_missing))
    # check if the file exists
    if (check_path_exists(path) == False):
        if (ignore_if_missing == False):
            raise Exception("delete_dir: path doesnt exist: {}".format(path))
        else:
            utils.debug("delete_dir: path doesnt exist: {}".format(path))

        return

    # delete
    os.rmdir(path)



# central place for all serialized version of the tsvs
from omigo_core import tsv, utils, timefuncs
from omigo_hydra import cluster_common_v2
from omigo_hydra import cluster_class_reflection
from omigo_hydra.cluster_common_v2 import ClusterCapabilities
import sys
from omigo_ext.splunk_ext import SplunkTSV
from omigo_hydra.cluster_shell_ext import SparkJobShellExecutorTSV
from omigo_core.tsv import TSV

SINGLE_SPLIT = 1
# awk command
# awk -F '(' '{print $1}'|awk -F' ' '{print "    def "$2"(self, *args, **kwargs):\n        return HydraTSV(self.header, self.data, super().ctx, super().__copy_and_append_operations__(cluster_common_v2.ClusterMapOperation(TSV."$2", self.requirements, *args, **kwargs))\n"}'|pbcopy
# TODO: None of the hydra class can return non TSV
# TODO: instead of taking Map Reduce for public methods, return the internal breakdown of omigo_core TSV apis
# its possible that the keys used for grouping are created later in some extension class
class HydraBaseTSV:
    def __init__(self, header, data):
        self.header = header
        self.data = data
        self.ctx = None
        self.operations = []
        self.num_splits = 1
        self.requirements = []

    def set_hydra_ctx(self, ctx):
        self.ctx = ctx
        return self

    def set_hydra_operations(self, operations):
        self.operations = operations
        return self

    def set_hydra_num_splits(self, num_splits):
        self.num_splits = num_splits
        return self

    def set_hydra_requirements(self, requirements):
        self.requirements = requirements
        return self

    def __get_columns__(self):
        return self.header.split("\t")

    def __copy_and_append_operations__(self, new_op):
        utils.debug("__copy_and_append_operations__: new_op: {}".format(new_op.name))
        # create new array
        new_operations = []
        for op in self.operations:
            new_operations.append(op)

        # append new operation
        new_operations.append(new_op)

        # return
        return new_operations

    # TODO: implement inline execution
    def collect(self, input_ids, output_ids, start_ts = None, use_full_data = False):
        # resolve start_ts as timestamp
        if (start_ts is not None):
            start_ts = timefuncs.datetime_to_utctimestamp(start_ts)

        # resolve context
        if (self.ctx is not None):
            wf_id = self.ctx.execute_jobs(self.__get_tsv__(), self.__get_jobs_operations__(), input_ids, output_ids, start_ts = start_ts, use_full_data = use_full_data)
            return wf_id
        else:
            raise Exception("HydraBaseTSV: collect(): ctx is None and in-memory execution is yet to be implemented")

    # TODO: single output
    def materialize(self, input_ids, output_ids, start_ts = None, use_full_data = False):
        return self.collect(input_ids, output_ids, start_ts = start_ts, use_full_data = use_full_data)

    def persist(self, path):
        raise Exception("TBD")

    def __get_tsv__(self):
        return tsv.TSV(self.header, self.data)

    def get_operations(self):
        return self.operations

    def __get_jobs_operations__(self):
        # check if there are any reduce operations
        reduce_indexes = []
        for i in range(len(self.operations)):
            # check for operation type
            if (isinstance(self.operations[i], (cluster_common_v2.ClusterReduceOperation, cluster_common_v2.ClusterSingletonOperation))):
                reduce_indexes.append(i)

        # debug
        utils.debug("HydraBaseTSV: __get_jobs_operations__: reduce operation indexes: {}".format(reduce_indexes))
        # for i in reduce_indexes:
        #     utils.debug("HydraBaseTSV: __get_jobs_operations__: reduce operation index: {}: {}".format(i, self.operations[i].name))

        # create a temporary split
        job_mr_splits = []

        # if there are no reduce operations, then a single job with only maps
        if (len(reduce_indexes) == 0):
            map_ops = self.operations
            reduce_op = None
            singleton_op = None
            job_mr_splits.append((map_ops, reduce_op, singleton_op))
        else:
            # iterate and add jobs
            for i in range(len(reduce_indexes)):
                # create placeholders
                map_ops = []
                reduce_op = None
                singleton_op = None
                cluster_def_op = None

                # boundary condition
                index = reduce_indexes[i]
                reduce_op = self.operations[index] if (isinstance(self.operations[index], (cluster_common_v2.ClusterReduceOperation))) else None
                singleton_op = self.operations[index] if (isinstance(self.operations[index], (cluster_common_v2.ClusterSingletonOperation))) else None

                # indexes
                if (i > 0):
                    prev_index = reduce_indexes[i-1]
                    map_ops = self.operations[prev_index+1:index]
                else:
                    map_ops = self.operations[0:index]

                # add to the job
                job_mr_splits.append((map_ops, reduce_op, singleton_op))

            # boundary condition if last sequence of operations was map only. TODO: dont use isinstance
            if (isinstance(self.operations[-1], (cluster_common_v2.ClusterReduceOperation), cluster_common_v2.ClusterSingletonOperation) == False):
                map_ops = self.operations[reduce_indexes[-1]+1:]
                reduce_op = None
                singleton_op = None
                job_mr_splits.append((map_ops, reduce_op, singleton_op))

        # create segments of map and reduce operations
        jobs_operations = []

        # job mr splits has map-reduce splits. Now split each further based on the ExtendClass
        for (map_ops, reduce_op, singleton_op) in job_mr_splits:
            # create indexes
            extend_class_indexes = []

            # check for None
            if (map_ops is not None and len(map_ops) > 0):
                # iterate
                for i in range(len(map_ops)):
                    # map_op
                    map_op = map_ops[i]

                    # check if this is extend class
                    if (isinstance(map_op, (cluster_common_v2.ClusterExtendClassOperation))):
                        extend_class_indexes.append(i)

                # debug
                utils.debug("HydraBaseTSV: __get_jobs_operations__: extend_class_op indexes: {}".format(extend_class_indexes))
                # for i in extend_class_indexes:
                #     utils.debug("HydraBaseTSV: __get_jobs_operations__: extend_class_op index: {}: {}".format(i, map_ops[i].name))

                # check if any indexes were found
                if (len(extend_class_indexes) > 0):
                    # iterate through indexes and split
                    for i in range(len(extend_class_indexes)):
                        cur_index = extend_class_indexes[i]

                        # the index is pointing to an extend class
                        class_def_op = map_ops[cur_index]

                        # first index is special
                        if (i == 0):
                            if (cur_index > 0):
                                map_ops2 = map_ops[0:cur_index]
                                reduce_op2 = None
                                singleton_op2 = None
                                jobs_operations.append(cluster_common_v2.ClusterOperationJob(map_ops2, reduce_op2, singleton_op2, None))

                        # check if it is the last split or before
                        if (i < len(extend_class_indexes) - 1):
                            # reduce_op is None for all but the last split
                            next_index = extend_class_indexes[i+1]
                            map_ops2 = map_ops[cur_index+1:next_index]
                            reduce_op2 = None
                            singleton_op2 = None
                            jobs_operations.append(cluster_common_v2.ClusterOperationJob(map_ops2, reduce_op2, singleton_op2, class_def_op))
                        else:
                            # last split gets the reduce
                            map_ops2 = map_ops[cur_index+1:]
                            reduce_op2 = reduce_op
                            singleton_op2 = singleton_op
                            jobs_operations.append(cluster_common_v2.ClusterOperationJob(map_ops2, reduce_op2, singleton_op2, class_def_op))
                else:
                    class_def_op = None
                    jobs_operations.append(cluster_common_v2.ClusterOperationJob(map_ops, reduce_op, singleton_op, class_def_op))
            else:
                class_def_op = None
                jobs_operations.append(cluster_common_v2.ClusterOperationJob(map_ops, reduce_op, singleton_op, class_def_op))

        # return
        return jobs_operations

# this is a shadow copy of TSV. Means APIs exist but TSV is not a base class.
class HydraTSV(HydraBaseTSV):
    def __init__(self, header, data, ctx = None, operations = [], requirements = []):
        super().__init__(header, data)
        super().set_hydra_ctx(ctx)
        super().set_hydra_operations(operations)
        super().set_hydra_requirements(requirements)

    def to_string(self, *args, **kwargs):
        raise Exception("Not implemented")

    def __new_hydra_tsv__(self, new_op):
        return HydraTSV(self.header, self.data, self.ctx, self.__copy_and_append_operations__(new_op))

    def to_string(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.to_string, self.requirements, *args, **kwargs))

    def get_content_as_string(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.get_content_as_string, self.requirements, *args, **kwargs))

    def validate(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.validate, self.requirements, *args, **kwargs))

    def has_col(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.has_col, self.requirements, *args, **kwargs))

    def select(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.select, self.requirements, *args, **kwargs))

    def values_not_in(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.values_not_in, self.requirements, *args, **kwargs))

    def values_in(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.values_in, self.requirements, *args, **kwargs))

    def not_match(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.not_match, self.requirements, *args, **kwargs))

    def not_regex_match(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.not_regex_match, self.requirements, *args, **kwargs))

    def match(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.match, self.requirements, *args, **kwargs))

    def regex_match(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.regex_match, self.requirements, *args, **kwargs))

    def not_eq(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.not_eq, self.requirements, *args, **kwargs))

    def eq(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.eq, self.requirements, *args, **kwargs))

    def eq_int(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.eq_int, self.requirements, *args, **kwargs))

    def eq_float(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.eq_float, self.requirements, *args, **kwargs))

    def eq_str(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.eq_str, self.requirements, *args, **kwargs))

    def not_eq_int(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.not_eq_int, self.requirements, *args, **kwargs))

    def not_eq_float(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.not_eq_float, self.requirements, *args, **kwargs))

    def not_eq_str(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.not_eq_str, self.requirements, *args, **kwargs))

    def is_nonzero(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.is_nonzero, self.requirements, *args, **kwargs))

    def is_nonzero_int(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.is_nonzero_int, self.requirements, *args, **kwargs))

    def is_nonzero_float(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.is_nonzero_float, self.requirements, *args, **kwargs))

    def lt_str(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.lt_str, self.requirements, *args, **kwargs))

    def le_str(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.le_str, self.requirements, *args, **kwargs))

    def gt_str(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.gt_str, self.requirements, *args, **kwargs))

    def ge_str(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.ge_str, self.requirements, *args, **kwargs))

    def gt(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.gt, self.requirements, *args, **kwargs))

    def gt_int(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.gt_int, self.requirements, *args, **kwargs))

    def gt_float(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.gt_float, self.requirements, *args, **kwargs))

    def ge(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.ge, self.requirements, *args, **kwargs))

    def ge_int(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.ge_int, self.requirements, *args, **kwargs))

    def ge_float(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.ge_float, self.requirements, *args, **kwargs))

    def lt(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.lt, self.requirements, *args, **kwargs))

    def lt_int(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.lt_int, self.requirements, *args, **kwargs))

    def lt_float(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.lt_float, self.requirements, *args, **kwargs))

    def le(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.le, self.requirements, *args, **kwargs))

    def le_int(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.le_int, self.requirements, *args, **kwargs))

    def le_float(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.le_float, self.requirements, *args, **kwargs))

    def startswith(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.startswith, self.requirements, *args, **kwargs))

    def not_startswith(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.not_startswith, self.requirements, *args, **kwargs))

    def endswith(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.endswith, self.requirements, *args, **kwargs))

    def not_endswith(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.not_endswith, self.requirements, *args, **kwargs))

    def is_empty_str(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.is_empty_str, self.requirements, *args, **kwargs))

    def is_nonempty_str(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.is_nonempty_str, self.requirements, *args, **kwargs))

    def replace_str_inline(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.replace_str_inline, self.requirements, *args, **kwargs))

    def group_count(self, *args, **kwargs):
        grouping_cols = args[0]
        op = cluster_common_v2.ClusterReduceOperation(grouping_cols, self.num_splits, TSV.group_count, self.requirements, *args, **kwargs)
        return self.__new_hydra_tsv__(op)

    def ratio(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.ratio, self.requirements, *args, **kwargs))

    def ratio_const(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.ratio_const, self.requirements, *args, **kwargs))

    def apply_precision(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.apply_precision, self.requirements, *args, **kwargs))

    def skip(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.skip, self.requirements, *args, **kwargs))

    def skip_rows(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.skip_rows, self.requirements, *args, **kwargs))

    def last(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.last, self.requirements, *args, **kwargs))

    def take(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.take, self.requirements, *args, **kwargs))

    def distinct(self, *args, **kwargs):
        grouping_cols = self.__get_columns__()
        op = cluster_common_v2.ClusterReduceOperation(grouping_cols, self.num_splits, TSV.distinct, self.requirements, *args, **kwargs)
        return self.__new_hydra_tsv__(op)

    def distinct_cols(self, *args, **kwargs):
        grouping_cols = self.__get_columns__()
        op = cluster_common_v2.ClusterReduceOperation(grouping_cols, self.num_splits, TSV.distinct_cols, self.requirements, *args, **kwargs)
        return self.__new_hydra_tsv__(op)

    def drop(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.drop, self.requirements, *args, **kwargs))

    def drop_cols(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.drop_cols, self.requirements, *args, **kwargs))

    def drop_cols_with_prefix(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.drop_cols_with_prefix, self.requirements, *args, **kwargs))

    def drop_cols_with_suffix(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.drop_cols_with_suffix, self.requirements, *args, **kwargs))

    def drop_if_exists(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.drop_if_exists, self.requirements, *args, **kwargs))

    def drop_cols_if_exists(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.drop_cols_if_exists, self.requirements, *args, **kwargs))

    def drop_empty_cols(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.drop_empty_cols, self.requirements, *args, **kwargs))

    def drop_empty_rows(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.drop_empty_rows, self.requirements, *args, **kwargs))

    def window_aggregate(self, *args, **kwargs):
        # create operation
        select_cols = kwargs["select_cols"] if ("select_cols" in kwargs.keys() and kwargs["select_cols"] is not None) else []

        # create grouping cols
        grouping_cols = []
        for c in select_cols:
            grouping_cols.append(c)
        grouping_cols.append(args[0])

    def group_by_key(self, *args, **kwargs):
        grouping_cols = args[0]
        op = cluster_common_v2.ClusterReduceOperation(grouping_cols, self.num_splits, TSV.group_by_key, self.requirements, *args, **kwargs)
        return self.__new_hydra_tsv__(op)

    def arg_min(self, *args, **kwargs):
        grouping_cols = args[0]
        op = cluster_common_v2.ClusterReduceOperation(grouping_cols, self.num_splits, TSV.arg_min, self.requirements, *args, **kwargs)
        return self.__new_hydra_tsv__(op)

    def arg_max(self, *args, **kwargs):
        grouping_cols = args[0]
        op = cluster_common_v2.ClusterReduceOperation(grouping_cols, self.num_splits, TSV.arg_max, self.requirements, *args, **kwargs)
        return self.__new_hydra_tsv__(op)

    def aggregate(self, *args, **kwargs):
        grouping_cols = args[0]
        op = cluster_common_v2.ClusterReduceOperation(grouping_cols, self.num_splits, TSV.aggregate, self.requirements, *args, **kwargs)
        return self.__new_hydra_tsv__(op)

    def filter(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.filter, self.requirements, *args, **kwargs))

    def exclude_filter(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.exclude_filter, self.requirements, *args, **kwargs))

    def any_col_with_cond_exists_filter(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.any_col_with_cond_exists_filter, self.requirements, *args, **kwargs))

    def any_col_with_cond_exists_exclude_filter(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.any_col_with_cond_exists_exclude_filter, self.requirements, *args, **kwargs))

    def all_col_with_cond_exists_filter(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.all_col_with_cond_exists_filter, self.requirements, *args, **kwargs))

    def all_col_with_cond_exists_exclude_filter(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.all_col_with_cond_exists_exclude_filter, self.requirements, *args, **kwargs))

    def transform(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.transform, self.requirements, *args, **kwargs))

    def transform_inline(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.transform_inline, self.requirements, *args, **kwargs))

    def transform_inline_log(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.transform_inline_log, self.requirements, *args, **kwargs))

    def transform_inline_log2(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.transform_inline_log2, self.requirements, *args, **kwargs))

    def transform_inline_log10(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.transform_inline_log10, self.requirements, *args, **kwargs))

    def transform_inline_log1p(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.transform_inline_log1p, self.requirements, *args, **kwargs))

    def transform_inline_log1p_base2(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.transform_inline_log1p_base2, self.requirements, *args, **kwargs))

    def transform_inline_log1p_base10(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.transform_inline_log1p_base10, self.requirements, *args, **kwargs))

    def rename(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.rename, self.requirements, *args, **kwargs))

    def get_header(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.get_header, self.requirements, *args, **kwargs))

    def get_data(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.get_data, self.requirements, *args, **kwargs))

    def get_header_map(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.get_header_map, self.requirements, *args, **kwargs))

    def num_rows(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.num_rows, self.requirements, *args, **kwargs))

    def num_cols(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.num_cols, self.requirements, *args, **kwargs))

    def get_size_in_bytes(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.get_size_in_bytes, self.requirements, *args, **kwargs))

    def size_in_bytes(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.size_in_bytes, self.requirements, *args, **kwargs))

    def size_in_mb(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.size_in_mb, self.requirements, *args, **kwargs))

    def size_in_gb(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.size_in_gb, self.requirements, *args, **kwargs))

    def get_header_fields(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.get_header_fields, self.requirements, *args, **kwargs))

    def get_columns(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.get_columns, self.requirements, *args, **kwargs))

    def get_column(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.get_column, self.requirements, *args, **kwargs))

    def columns(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.columns, self.requirements, *args, **kwargs))

    def get_column_index(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.get_column_index, self.requirements, *args, **kwargs))

    def export_to_maps(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.export_to_maps, self.requirements, *args, **kwargs))

    def to_maps(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.to_maps, self.requirements, *args, **kwargs))

    def to_int(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.to_int, self.requirements, *args, **kwargs))

    def to_numeric(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.to_numeric, self.requirements, *args, **kwargs))

    def add_seq_num(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.add_seq_num, self.requirements, *args, **kwargs))

    def show_transpose(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.show_transpose, self.requirements, *args, **kwargs))

    def show(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.show, self.requirements, *args, **kwargs))

    def show_sample(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.show_sample, self.requirements, *args, **kwargs))

    def col_as_array(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.col_as_array, self.requirements, *args, **kwargs))

    def col_as_float_array(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.col_as_float_array, self.requirements, *args, **kwargs))

    def col_as_int_array(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.col_as_int_array, self.requirements, *args, **kwargs))

    def col_as_array_uniq(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.col_as_array_uniq, self.requirements, *args, **kwargs))

    def col_as_array_uniq_non_empty(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.col_as_array_uniq_non_empty, self.requirements, *args, **kwargs))

    def cols_as_map(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.cols_as_map, self.requirements, *args, **kwargs))

    def sort(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.sort, self.requirements, *args, **kwargs))

    def reverse_sort(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.reverse_sort, self.requirements, *args, **kwargs))

    def numerical_sort(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.numerical_sort, self.requirements, *args, **kwargs))

    def reverse_numerical_sort(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.reverse_numerical_sort, self.requirements, *args, **kwargs))

    def reorder(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.reorder, self.requirements, *args, **kwargs))

    def reorder_reverse(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.reorder_reverse, self.requirements, *args, **kwargs))

    def reverse_reorder(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.reverse_reorder, self.requirements, *args, **kwargs))

    def noop(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.noop, self.requirements, *args, **kwargs))

    def to_df(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.to_df, self.requirements, *args, **kwargs))

    def to_simple_df(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.to_simple_df, self.requirements, *args, **kwargs))

    def export_to_df(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.export_to_df, self.requirements, *args, **kwargs))

    def to_json_records(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.to_json_records, self.requirements, *args, **kwargs))

    def to_csv(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.to_csv, self.requirements, *args, **kwargs))

    def url_encode_inline(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.url_encode_inline, self.requirements, *args, **kwargs))

    def url_decode_inline(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.url_decode_inline, self.requirements, *args, **kwargs))

    def url_encode(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.url_encode, self.requirements, *args, **kwargs))

    def url_decode(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.url_decode, self.requirements, *args, **kwargs))

    def resolve_url_encoded_cols(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.resolve_url_encoded_cols, self.requirements, *args, **kwargs))

    def resolve_url_encoded_list_cols(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.resolve_url_encoded_list_cols, self.requirements, *args, **kwargs))

    def resolve_all_url_encoded_cols(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.resolve_all_url_encoded_cols, self.requirements, *args, **kwargs))

    def union(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.union, self.requirements, *args, **kwargs))

    def difference(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.difference, self.requirements, *args, **kwargs))

    def add_const(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.add_const, self.requirements, *args, **kwargs))

    def add_const_if_missing(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.add_const_if_missing, self.requirements, *args, **kwargs))

    def add_empty_cols_if_missing(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.add_empty_cols_if_missing, self.requirements, *args, **kwargs))

    def add_row(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.add_row, self.requirements, *args, **kwargs))

    def add_map_as_row(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.add_map_as_row, self.requirements, *args, **kwargs))

    def assign_value(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.assign_value, self.requirements, *args, **kwargs))

    def concat_as_cols(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.concat_as_cols, self.requirements, *args, **kwargs))

    def add_col_prefix(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.add_col_prefix, self.requirements, *args, **kwargs))

    def remove_suffix(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.remove_suffix, self.requirements, *args, **kwargs))

    def add_prefix(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.add_prefix, self.requirements, *args, **kwargs))

    def add_suffix(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.add_suffix, self.requirements, *args, **kwargs))

    def rename_prefix(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.rename_prefix, self.requirements, *args, **kwargs))

    def rename_suffix(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.rename_suffix, self.requirements, *args, **kwargs))

    def remove_prefix(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.remove_prefix, self.requirements, *args, **kwargs))

    def replace_prefix(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.replace_prefix, self.requirements, *args, **kwargs))

    def replace_suffix(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.replace_suffix, self.requirements, *args, **kwargs))

    def sample(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.sample, self.requirements, *args, **kwargs))

    def sample_without_replacement(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.sample_without_replacement, self.requirements, *args, **kwargs))

    def sample_with_replacement(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.sample_with_replacement, self.requirements, *args, **kwargs))

    def sample_rows(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.sample_rows, self.requirements, *args, **kwargs))

    def sample_n(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.sample_n, self.requirements, *args, **kwargs))

    def sample_n_with_warn(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.sample_n_with_warn, self.requirements, *args, **kwargs))

    def sample_n_with_replacement(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.sample_n_with_replacement, self.requirements, *args, **kwargs))

    def sample_n_without_replacement(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.sample_n_without_replacement, self.requirements, *args, **kwargs))

    def sample_group_by_topk_if_reached_limit(self, *args, **kwargs):
        limit = args[0]
        grouping_cols = args[1]
        op = cluster_common_v2.ClusterReduceOperation(grouping_cols, self.num_splits, TSV.sample_group_by_topk_if_reached_limit, self.requirements, *args, **kwargs)
        return self.__new_hydra_tsv__(op)

    def warn_if_limit_reached(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.warn_if_limit_reached, self.requirements, *args, **kwargs))

    def cap_min_inline(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.cap_min_inline, self.requirements, *args, **kwargs))

    def cap_max_inline(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.cap_max_inline, self.requirements, *args, **kwargs))

    def cap_min(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.cap_min, self.requirements, *args, **kwargs))

    def cap_max(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.cap_max, self.requirements, *args, **kwargs))

    def copy(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.copy, self.requirements, *args, **kwargs))

    def sample_class(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.sample_class, self.requirements, *args, **kwargs))

    def sample_group_by_col_value(self, *args, **kwargs):
        grouping_cols = args[0]
        op = cluster_common_v2.ClusterReduceOperation(grouping_cols, self.num_splits, TSV.sample_group_by_col_value, self.requirements, *args, **kwargs)
        return self.__new_hydra_tsv__(op)

    def sample_group_by_max_uniq_values_exact(self, *args, **kwargs):
        grouping_cols = args[0]
        op = cluster_common_v2.ClusterReduceOperation(grouping_cols, self.num_splits, TSV.sample_group_by_max_uniq_values_exact, self.requirements, *args, **kwargs)
        return self.__new_hydra_tsv__(op)

    def sample_group_by_max_uniq_values_approx(self, *args, **kwargs):
        grouping_cols = args[0]
        op = cluster_common_v2.ClusterReduceOperation(grouping_cols, self.num_splits, TSV.sample_group_by_max_uniq_values_approx, self.requirements, *args, **kwargs)
        return self.__new_hydra_tsv__(op)

    def sample_group_by_max_uniq_values(self, *args, **kwargs):
        grouping_cols = args[0]
        op = cluster_common_v2.ClusterReduceOperation(grouping_cols, self.num_splits, TSV.sample_group_by_max_uniq_values, self.requirements, *args, **kwargs)
        return self.__new_hydra_tsv__(op)

    def sample_group_by_max_uniq_values_per_class(self, *args, **kwargs):
        grouping_cols = args[0]
        op = cluster_common_v2.ClusterReduceOperation(grouping_cols, self.num_splits, TSV.sample_group_by_max_uniq_values_per_class, self.requirements, *args, **kwargs)
        return self.__new_hydra_tsv__(op)

    def sample_group_by_key(self, *args, **kwargs):
        grouping_cols = args[0]
        op = cluster_common_v2.ClusterReduceOperation(grouping_cols, self.num_splits, TSV.sample_group_by_key, self.requirements, *args, **kwargs)
        return self.__new_hydra_tsv__(op)

    def sample_column_by_max_uniq_values(self, *args, **kwargs):
        grouping_cols = args[0]
        op = cluster_common_v2.ClusterReduceOperation(grouping_cols, self.num_splits, TSV.sample_column_by_max_uniq_values, self.requirements, *args, **kwargs)
        return self.__new_hydra_tsv__(op)

    def sample_class_by_min_class_count(self, *args, **kwargs):
        grouping_cols = args[0]
        op = cluster_common_v2.ClusterReduceOperation(grouping_cols, self.num_splits, TSV.sample_class_by_min_class_count, self.requirements, *args, **kwargs)
        return self.__new_hydra_tsv__(op)

    def sample_class_by_max_values(self, *args, **kwargs):
        grouping_cols = args[0]
        op = cluster_common_v2.ClusterReduceOperation(grouping_cols, self.num_splits, TSV.sample_class_by_max_values, self.requirements, *args, **kwargs)
        return self.__new_hydra_tsv__(op)

    def left_join(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.left_join, self.requirements, *args, **kwargs))

    def right_join(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.right_join, self.requirements, *args, **kwargs))

    def inner_join(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.inner_join, self.requirements, *args, **kwargs))

    def outer_join(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.outer_join, self.requirements, *args, **kwargs))

    def join(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.join, self.requirements, *args, **kwargs))

    def natural_join(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.natural_join, self.requirements, *args, **kwargs))

    def inner_map_join(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.inner_map_join, self.requirements, *args, **kwargs))

    def left_map_join(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.left_map_join, self.requirements, *args, **kwargs))

    def split_batches(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.split_batches, self.requirements, *args, **kwargs))

    def generate_key_hash(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.generate_key_hash, self.requirements, *args, **kwargs))

    def cumulative_sum(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.cumulative_sum, self.requirements, *args, **kwargs))

    def replicate_rows(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.replicate_rows, self.requirements, *args, **kwargs))

    def explode(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.explode, self.requirements, *args, **kwargs))

    def explode_json(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.explode_json, self.requirements, *args, **kwargs))

    def explode_json_v2(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.explode_json_v2, self.requirements, *args, **kwargs))

    def transpose(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.transpose, self.requirements, *args, **kwargs))

    def reverse_transpose(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.reverse_transpose, self.requirements, *args, **kwargs))

    def flatmap(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.flatmap, self.requirements, *args, **kwargs))

    def to_tuples(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.to_tuples, self.requirements, *args, **kwargs))

    def set_missing_values(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.set_missing_values, self.requirements, *args, **kwargs))

    def extend_class(self, *args, **kwargs):
        # find the corresponding hydra version of the class.
        hydra_class_ref = cluster_class_reflection.get_hydra_class(newclass, [sys.modules[__name__]])

        # check for None
        if (hydra_class_ref is None):
            raise Exception("HydraTSV: extend_class: not able to find hydra version for class: {}".format(newclass.__name__))

        # add the parameters
        hydra_class_obj = hydra_class_ref(self.header, self.data)
        hydra_class_obj.set_hydra_ctx(self.ctx)
        hydra_class_obj.set_hydra_operations(self.__copy_and_append_operations__(cluster_common_v2.ClusterExtendClassOperation(newclass, hydra_class_obj.requirements, *args, **kwargs)))

        # return
        return hydra_class_obj

    def extend_external_class(self, *args, **kwargs):
        raise Exception("Not implemented")

    def custom_func(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.custom_func, self.requirements, *args, **kwargs))

    def to_clipboard(self, *args, **kwargs):
        raise Exception("Not implemented")

    def filter_json_by_xpath(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.filter_json_by_xpath, self.requirements, *args, **kwargs))

    def get_col_index(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.get_col_index, self.requirements, *args, **kwargs))

    def get_hash(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.get_hash, self.requirements, *args, **kwargs))

    def is_empty(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.is_empty, self.requirements, *args, **kwargs))

    def has_empty_header(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.has_empty_header, self.requirements, *args, **kwargs))

    def write(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.write, self.requirements, *args, **kwargs))

    def show_custom_func(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.show_custom_func, self.requirements, *args, **kwargs))

    def show_group_count(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.show_group_count, self.requirements, *args, **kwargs))

    def show_select_func(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.show_select_func, self.requirements, *args, **kwargs))

    def show_transpose_custom_func(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.show_transpose_custom_func, self.requirements, *args, **kwargs))

    def print(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.print, self.requirements, *args, **kwargs))

    def print_stats(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.print_stats, self.requirements, *args, **kwargs))

    def warn(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.warn, self.requirements, *args, **kwargs))

    def warn_once(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.warn_once, self.requirements, *args, **kwargs))

    def get_max_size_cols_stats(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterSingletonOperation(TSV.get_max_size_cols_stats, self.requirements, *args, **kwargs))

    def sleep(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.sleep, self.requirements, *args, **kwargs))

    def split(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.split, self.requirements, *args, **kwargs))

    def split_str(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.split_str, self.requirements, *args, **kwargs))

    def sample_group_by_topk(self, *args, **kwargs):
        grouping_cols = args[0]
        op = cluster_common_v2.ClusterReduceOperation(grouping_cols, self.num_splits, TSV.sample_group_by_topk, self.requirements, *args, **kwargs)
        return self.__new_hydra_tsv__(op)

    def resolve_template_col(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.resolve_template_col, self.requirements, *args, **kwargs))

    def resolve_template_col_inline(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.resolve_template_col_inline, self.requirements, *args, **kwargs))

    def enable_info_mode(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.enable_info_mode, self.requirements, *args, **kwargs))

    def disable_info_mode(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.disable_info_mode, self.requirements, *args, **kwargs))

    def enable_debug_mode(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.enable_debug_mode, self.requirements, *args, **kwargs))

    def disable_debug_mode(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.disable_debug_mode, self.requirements, *args, **kwargs))

    def enable_trace_mode(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.enable_trace_mode, self.requirements, *args, **kwargs))

    def disable_trace_mode(self, *args, **kwargs):
        return self.__new_hydra_tsv__(cluster_common_v2.ClusterMapOperation(TSV.disable_trace_mode, self.requirements, *args, **kwargs))

class HydraHelper:
    def new_hydra_tsv(hydra_base, new_op):
        return HydraTSV(hydra_base.header, hydra_base.data, ctx = hydra_base.ctx, operations = hydra_base.__copy_and_append_operations__(new_op))

class HydraSparkJobShellExecutorTSV(HydraBaseTSV):
    def __init__(self, header, data):
        super().__init__(header, data)
        super().set_hydra_requirements([ClusterCapabilities.SPARK, ClusterCapabilities.SHELL])

    def execute(self, *args, **kwargs):
        return HydraHelper.new_hydra_tsv(self, cluster_common_v2.ClusterMapOperation(SparkJobShellExecutorTSV.execute, self.requirements, *args, **kwargs))

class HydraSplunkTSV(HydraBaseTSV):
    def __init__(self, header, data):
        super().__init__(header, data)
        super().set_hydra_requirements([ClusterCapabilities.SPLUNK])

    def get_events_par(self, *args, **kwargs):
        return HydraHelper.new_hydra_tsv(self, cluster_common_v2.ClusterMapOperation(SplunkTSV.get_events_par, self.requirements, *args, **kwargs))

    def get_events_parsed(self, *args, **kwargs):
        return HydraHelper.new_hydra_tsv(self, cluster_common_v2.ClusterMapOperation(SplunkTSV.get_events_parsed, self.requirements, *args, **kwargs))




"""FilePathsDataReader class"""

import sys

# local import
from omigo_hydra import file_paths_reader
from omigo_hydra import file_paths_util

class FilePathsDataReader:
    """A simple class to read filepaths data"""

    s3_region = ""
    aws_profile = ""

    # file paths reader is an iterator based reader
    file_paths_readers = None

    # cur_data holds the current data block
    cur_data = None

    # cur_index is the pointer to the line in cur_data
    cur_index = 0

    # header is the header line for entire dataset
    header = None

    # constructor
    def __init__(self, filepaths, s3_region, aws_profile):
        # s3 config
        self.s3_region = s3_region
        self.aws_profile = aws_profile

        # initialize the reader
        self.file_paths_readers = file_paths_reader.FilePathsReader(filepaths)

        # loop through all files until find a one with proper data
        found = False
        while (found == False):
            if (self.file_paths_readers.has_next()):
                # initial load
                filepath = self.file_paths_readers.next()
                self.cur_data = file_paths_util.read_file_content_as_lines(filepath, self.s3_region, self.aws_profile)

                # check for validity
                if (len(self.cur_data) == 0):
                    print("FilePathsDataReader: Invalid file found. No header.", filepath)
                    sys.exit(0)

                # read header
                self.header = self.cur_data[0].rstrip("\n")

                # assign cur_index to 0 for reading header
                self.cur_index = 1

                # check if found a valid file
                if (self.cur_index < len(self.cur_data)):
                    found = True
            else:
                self.cur_data = None
                self.cur_index = -1
                break

        # check if found some valid file to read
        if (found == False):
            self.cur_data = None
            self.cur_index = -1

    # get header
    def get_header(self):
        return self.header

    # has next returns boolean if there is still some data left
    def has_next(self):
        # check if no data block loaded
        if (self.cur_data is None):
            return False

        return True

    # close the file reader
    def close(self):
        self.cur_data = None
        self.cur_index = -1

    # next return the next line and also moves the pointers
    def next(self):
        # declare and initialize variable
        result = None

        # cur_data should be non empty
        if (self.has_next() == False):
            print("FilePathsDataReader: next() has next is false and next is called")
            return None

        # check for current pointer
        if (self.cur_index < len(self.cur_data)):
            result = self.cur_data[self.cur_index].rstrip("\n")
            self.cur_index = self.cur_index + 1
        else:
            print("FilePathsDataReader: next() this should not happen")

        # Check if cur_index has reached the end of data block
        if (self.cur_index >= len(self.cur_data)):
            # check if any more data blocks are left
            if (self.file_paths_readers.has_next() == False):
                self.cur_data = None
                self.cur_index = -1
            else:
                found = False
                while (found == False):
                    if (self.file_paths_readers.has_next()):
                        # read the next block and reinitialize the cur_index
                        filepath = self.file_paths_readers.next()
                        self.cur_data = file_paths_util.read_file_content_as_lines(filepath, self.s3_region, self.aws_profile)
                        self.cur_index = 1

                        # check if found a valid file
                        if (self.cur_index < len(self.cur_data)):
                            found = True
                    else:
                        # invalidate the buffers
                        self.cur_data = None
                        self.cur_index = -1
                        break

        # return result
        return result




from omigo_core import dataframe, utils, timefuncs
from omigo_hydra import s3io_wrapper
import json
import time
import base64
import sys
import traceback

# Increment this version to rerun analysis
RUN_VERSION = "version:v01"
DEFAULT_WAIT_10SEC = 10
DEFAULT_WAIT_30SEC = 30
DEFAULT_WAIT_60SEC = 60
DEFAULT_RETRY_ATTEMPTS = 3
DEFAULT_RETRY_WAIT_SEC = 60

# Class RunnerBase
class RunnerBase:
    def __init__(self, base_dir):
        self.base_dir = base_dir
        self.fs = s3io_wrapper.S3FSWrapper()

        # debug
        utils.info("RunnerBase: init: base_dir: {}".format(self.base_dir))

    def get_base_dir(self):
        return self.base_dir

    def get_output_dir(self):
        return "{}/output".format(self.base_dir)

    def get_job_status_dir(self):
        return "{}/job-status".format(self.base_dir)

    def get_job_tags_dir(self):
        return "{}/job-tags".format(self.base_dir)

    def get_job_status_created(self):
        return "{}/created".format(self.get_job_status_dir())

    def get_job_status_running(self):
        return "{}/running".format(self.get_job_status_dir())

    def get_job_status_completed(self):
        return "{}/completed".format(self.get_job_status_dir())

    def get_job_status_created_job_id_file(self, job_id):
        return "{}/{}".format(self.get_job_status_created(), job_id)

    def get_job_status_running_job_id_file(self, job_id):
        return "{}/{}".format(self.get_job_status_running(), job_id)

    def get_job_status_completed_job_id_file(self, job_id):
        return "{}/{}".format(self.get_job_status_completed(), job_id)

    def get_workers_dir(self):
        return "{}/workers".format(self.base_dir)

    def get_worker_dir(self, worker_id):
        return "{}/{}".format(self.get_workers_dir(), worker_id)

    def get_worker_jobs_dir(self, worker_id):
        return "{}/jobs".format(self.get_worker_dir(worker_id))

    def get_worker_job_id_file(self, worker_id, job_id):
        return "{}/{}".format(self.get_worker_jobs_dir(worker_id), job_id)

    def get_job_output_dir(self, job_id):
        return "{}/{}".format(self.get_output_dir(), job_id)

    def get_job_graph_analysis_file(self, job_id):
        return "{}/graph_analysis.tsv.gz".format(self.get_job_output_dir(job_id))

    def get_job_progressive_output_file(self, job_id):
        return "{}/progressive_output.txt".format(self.get_job_output_dir(job_id))

    def get_job_final_output_file(self, job_id):
        return "{}/final_output.txt".format(self.get_job_output_dir(job_id))

    def get_job_tag_empty_dir(self, job_id):
        return "{}/{}".format(self.get_job_tags_dir(), "empty")

    def get_job_tag_empty_dir_by_job_id(self, job_id):
        return "{}/{}".format(self.get_job_tag_empty_dir(), job_id)

    def get_job_tag_combined_dir(self):
        return "{}/{}".format(self.get_job_tags_dir(), "combined")

    def get_job_tag_empty_dir_by_job_id(self, job_id):
        return "{}/{}".format(self.get_job_tag_empty_dir(), job_id)

    def get_job_tag_combined_dir_by_job_id(self, job_id, key, value):
        value_b16 = base64.b16encode(bytes(str(value), "utf-8")).decode("utf-8")
        return "{}/{}-{}={}".format(self.get_job_tag_combined_dir(), job_id, key, value_b16)

    def get_shutdown_file(self):
        return "{}/shutdown".format(self.base_dir)

    def get_halt_file(self):
        return "{}/halt".format(self.base_dir)

    def is_shutdown_mode(self):
        return self.fs.file_exists(self.get_shutdown_file())

# class ClusterAdmin
class ClusterAdmin:
    def __init__(self, base_dir, wait_sec = DEFAULT_WAIT_60SEC):
        self.runner_base = RunnerBase(base_dir)
        self.wait_sec = wait_sec

    # create cluster
    def create_cluster(self, num_workers = 3):
        # print the cluster base path
        utils.info("create_cluster: Using base path: {}".format(self.runner_base.get_base_dir()))

        # create all paths
        self.runner_base.fs.create_dir(self.runner_base.get_base_dir())
        self.runner_base.fs.create_dir(self.runner_base.get_output_dir())
        self.runner_base.fs.create_dir(self.runner_base.get_job_status_dir())
        self.runner_base.fs.create_dir(self.runner_base.get_job_tags_dir())
        self.runner_base.fs.create_dir(self.runner_base.get_job_status_created())
        self.runner_base.fs.create_dir(self.runner_base.get_job_status_running())
        self.runner_base.fs.create_dir(self.runner_base.get_job_status_completed())
        self.runner_base.fs.create_dir(self.runner_base.get_workers_dir())
        self.runner_base.fs.create_dir(self.runner_base.get_job_tag_combined_dir())

        # create workers
        self.create_workers(num_workers)

    # create workers
    def create_workers(self, num_workers):
        # guards
        if (num_workers >= 100 or num_workers < 0):
            raise Exception("create_workers: number of workers need to be below 100: {}".format(num_workers))

        # run a loop
        for i in range(num_workers):
            worker_id = "worker{:02d}".format(i + 1)
            self.runner_base.fs.create_dir(self.runner_base.get_worker_dir(worker_id))
            self.runner_base.fs.create_dir(self.runner_base.get_worker_jobs_dir(worker_id))

    # delete job
    def delete_job(self, job_id, ignore_if_missing = True):
        # delete job tags
        for filename in self.runner_base.fs.list_leaf_dir(self.runner_base.get_job_tag_combined_dir()):
            if (filename.startswith("{}-".format(job_id)) == True):
                self.runner_base.fs.delete_file("{}/{}".format(self.runner_base.get_job_tag_combined_dir(), filename), ignore_if_missing = ignore_if_missing)

        # delete job output dir
        if (self.runner_base.fs.dir_exists(self.runner_base.get_job_output_dir(job_id))):
            for filename in self.runner_base.fs.list_leaf_dir(self.runner_base.get_job_output_dir(job_id)):
                self.runner_base.fs.delete_file("{}/{}".format(self.runner_base.get_job_output_dir(job_id), filename), ignore_if_missing = ignore_if_missing)

            # delete directory
            self.runner_base.fs.delete_dir_with_wait(self.runner_base.get_job_output_dir(job_id), ignore_if_missing = ignore_if_missing)

        # delete job-status
        self.runner_base.fs.delete_file(self.runner_base.get_job_status_completed_job_id_file(job_id), ignore_if_missing = ignore_if_missing)
        self.runner_base.fs.delete_file(self.runner_base.get_job_status_running_job_id_file(job_id), ignore_if_missing = ignore_if_missing)
        self.runner_base.fs.delete_file(self.runner_base.get_job_status_created_job_id_file(job_id), ignore_if_missing = ignore_if_missing)

        # goto each worker and delete the entry for job if exists
        for worker_id in self.runner_base.fs.list_dirs(self.runner_base.get_workers_dir()):
            if (self.runner_base.fs.file_exists(self.runner_base.get_worker_job_id_file(worker_id, job_id)) == True):
                utils.warn("Found job meant for deletion: {} running under worker: {}".format(job_id, worker_id))
                self.runner_base.fs.delete_file(self.runner_base.get_worker_job_id_file(worker_id, job_id), ignore_if_missing = ignore_if_missing)

    def resubmit(self, job_id, ignore_if_missing = True):
        # delete job-status
        self.runner_base.fs.delete_file(self.runner_base.get_job_status_completed_job_id_file(job_id), ignore_if_missing = ignore_if_missing)
        self.runner_base.fs.delete_file(self.runner_base.get_job_status_running_job_id_file(job_id), ignore_if_missing = ignore_if_missing)
        self.runner_base.fs.delete_file(self.runner_base.get_job_status_created_job_id_file(job_id), ignore_if_missing = ignore_if_missing)

# Class BaseJobGenerator
class BaseJobGenerator:
    def __init__(self, base_dir, wait_sec = DEFAULT_WAIT_60SEC, num_iter = 10):
        self.runner_base = RunnerBase(base_dir)
        self.wait_sec = wait_sec
        self.num_iter = num_iter

        # debug
        utils.info("BaseJobGenerator: base_dir: {}, wait_sec: {}, num_iter: {}".format(base_dir, wait_sec, num_iter))

    # run the jobs in a loop
    def run_loop(self):
        try:
            for i in range(self.num_iter):
                # check for shutdown mode
                if (self.runner_base.is_shutdown_mode() == True):
                    utils.warn("BaseJobGenerator: shutdown mode detected. Stopping")
                    return

                # run
                self.run()

                # sleep
                utils.info("BaseJobGenerator: Iteration: {}, Sleeping for {} seconds".format(i, self.wait_sec))
                time.sleep(self.wait_sec)
        except Exception as e:
            utils.error("BaseJobGenerator: run_loop: caught exception: {}. Returning".format(e))
            return
        except:
            utils.error("BaseJobGenerator: run_loop: caught interrupt. Returning")
            return

    # implement this method
    def run(self):
        raise Exception("Not implemented in base class")

# Class JobManager
class JobManager:
    def __init__(self, base_dir, wait_sec = DEFAULT_WAIT_30SEC):
        self.runner_base = RunnerBase(base_dir)
        self.wait_sec = wait_sec
        self.max_attempts = 3

        # debug
        utils.info("JobManager: base_dir: {}, wait_sec: {}".format(base_dir, wait_sec))

    def run_loop(self):
        num_attempts = 0
        # run for given number of attempts
        while (num_attempts < self.max_attempts):
            try:
              while True:
                  # check for shutdown mode
                  if (self.runner_base.is_shutdown_mode() == True):
                      utils.warn("JobManager: shutdown mode detected. Stopping")
                      return

                  # run
                  self.run()

                  # sleep
                  utils.info("JobManager: Sleeping for {} seconds".format(self.wait_sec))
                  time.sleep(self.wait_sec)
            except Exception as e:
                utils.error("JobManager: run_loop: caught exception: {}. attempts: {} / {}".format(e, num_attempts, self.max_attempts))
                # increase
                num_attempts = num_attempts + 1
            except:
                utils.error("JobManager: run_loop: caught Interrupt: returning")
                return

    def run(self):
        # check the list of created, running, completed
        all_created = self.runner_base.fs.list_leaf_dir(self.runner_base.get_job_status_created())
        all_running = self.runner_base.fs.list_leaf_dir(self.runner_base.get_job_status_running())
        all_completed = self.runner_base.fs.list_leaf_dir(self.runner_base.get_job_status_completed())

        # dedup sets to find the actual jobs that need to be run
        new_job_ids_created = list(filter(lambda t: t not in all_running and t not in all_completed, all_created))

        # iterate
        job_xdfs = [dataframe.new_with_cols(["job_id", "submit_ts"])]
        for new_job_id in new_job_ids_created:
            job_spec_json = json.loads(self.runner_base.fs.read_file_contents_as_text(self.runner_base.get_job_status_created_job_id_file(new_job_id)))
            job_xdfs.append(dataframe.from_maps([{"job_id": job_spec_json["job_id"], "submit_ts": job_spec_json["submit_ts"]}]))

        # create single tsv
        job_xdf = dataframe.merge_union(job_xdfs)

        # iterate
        for new_job_id in job_xdf.sort("submit_ts").col_as_array("job_id"):
            # debug
            utils.info("Looking for worker for job_id: {}".format(new_job_id))

            # read job spec
            job_spec_json = json.loads(self.runner_base.fs.read_file_contents_as_text(self.runner_base.get_job_status_created_job_id_file(new_job_id)))

            # find the worker that is not running anything
            worker_ids = self.runner_base.fs.list_leaf_dir(self.runner_base.get_workers_dir())

            # iterate and find idle worker
            idle_workers = []
            for worker_id in worker_ids:
                # get the list of current jobs taken by the worker
                worker_pending_jobs = self.runner_base.fs.list_leaf_dir(self.runner_base.get_worker_jobs_dir(worker_id))

                # check if the worker is idle
                if (len(worker_pending_jobs) == 0):
                    idle_workers.append(worker_id)

            # debug
            utils.info("Idle workers: {}".format(idle_workers))

            # put the new job id in the worker queue
            if (len(idle_workers) > 0):
                worker_id = idle_workers[0]
                job_spec_json["start_ts"] = timefuncs.utctimestamp_to_datetime_str(timefuncs.get_utctimestamp_sec())
                self.runner_base.fs.write_text_file(self.runner_base.get_worker_job_id_file(worker_id, new_job_id), "")
                self.runner_base.fs.write_text_file(self.runner_base.get_job_status_running_job_id_file(new_job_id), json.dumps(job_spec_json))
                utils.info("Job Assigned: job_id: {}: worker: {}".format(new_job_id, worker_id))

# Class ReprocessBatch
class ReprocessBatchWorker:
    def __init__(self, base_dir):
        self.runner_base = RunnerBase(base_dir)

        # debug
        utils.info("ReprocessBatchWorker: base_dir: {}".format(base_dir))

    def resubmit_analysis(self, last_n, version_tag = "version:v0"):
        # get job ids
        job_ids = sorted(self.runner_base.fs.list_leaf_dir(self.runner_base.get_job_status_completed()), reverse = True)

        # check for last_n
        job_ids = job_ids[0:last_n]

        # call rerun_specific_jobs
        for job_id in job_ids:
            self.resubmit_specific_analysis(job_id, version_tag)

    def resubmit_specific_analysis(self, job_id, version_tag):
        utils.info("ReprocessBatchWorker: resubmit_specific_analysis: {}".format(job_id))
        analysis_file = self.runner_base.get_job_graph_analysis_file(job_id)
        if (tsv.exists(analysis_file)):
            # read old file, drop previous version
            tsv.read(analysis_file) \
                .drop_cols_with_prefix("version", ignore_if_missing = True) \
                .drop_cols_if_exists("v0") \
                .add_const(version_tag, "1") \
                .write(analysis_file)

    def resubmit_completed(self, last_n, use_checkpoint = True):
        # add some guards
        if (last_n > 10):
            raise Exception("resubmit_completed: this moves the job status to CREATED. Not designed to be used for lot of requests")

        # get job ids
        job_ids = sorted(self.runner_base.fs.list_leaf_dir(self.runner_base.get_job_status_completed()), reverse = True)

        # check for last_n
        job_ids = job_ids[0:last_n]

        # call rerun_specific_jobs
        self.resubmit_specific_jobs(job_ids, use_checkpoint = use_checkpoint)

    def resubmit_created(self, last_n, use_checkpoint = True):
        # check for shutdown mode
        if (self.runner_base.is_shutdown_mode() == False):
            raise Exception("resubmit_created: This can only be called during shutdown mode")

        # add some guards
        if (last_n > 100):
            raise Exception("resubmit_created: this moves the job status to CREATED. Not designed to be used for lot of requests")

        # get job ids
        created_job_ids = self.runner_base.fs.list_leaf_dir(self.runner_base.get_job_status_created())
        running_job_ids = self.runner_base.fs.list_leaf_dir(self.runner_base.get_job_status_running())
        completed_job_ids = self.runner_base.fs.list_leaf_dir(self.runner_base.get_job_status_completed())
        job_ids = sorted(set(created_job_ids).difference(set(running_job_ids + completed_job_ids)), reverse = True)

        # check for last_n
        job_ids = job_ids[0:last_n]

        # call rerun_specific_jobs
        self.resubmit_specific_jobs(job_ids, use_checkpoint = use_checkpoint)

    def resubmit_failed(self, last_n, use_checkpoint = True):
        job_ids = sorted(self.runner_base.fs.list_leaf_dir(self.runner_base.get_job_status_completed()), reverse = True)

        # take only last n
        job_ids = job_ids[0:last_n]

        for job_id in job_ids:
            # read job spec
            job_spec_json = json.loads(self.runner_base.fs.read_file_contents_as_text(self.runner_base.get_job_status_completed_job_id_file(job_id)))

            # check if the status=FAILED
            if (job_spec_json["status"] == "FAILED"):
                utils.info("ReprocessBatchWorker: rerun_failed: failed job: {}".format(job_id))
                self.resubmit_specific_jobs([job_id], use_checkpoint = use_checkpoint)

    def resubmit_specific_jobs(self, job_ids, use_checkpoint = True):
        # iterate
        for job_id in job_ids:
            # read job spec
            job_spec_json = json.loads(self.runner_base.fs.read_file_contents_as_text(self.runner_base.get_job_status_created_job_id_file(job_id)))

            # change the status
            utils.info("ReprocessBatchWorker: resubmit_specific_jobs: job_id: {}".format(job_id))
            self.runner_base.fs.delete_file(self.runner_base.get_job_status_completed_job_id_file(job_id), ignore_if_missing = True)
            self.runner_base.fs.delete_file(self.runner_base.get_job_status_running_job_id_file(job_id), ignore_if_missing = True)

            # delete existing files if use_checkpoint = False
            if (use_checkpoint == False):
                # check if the directory exists
                if (self.runner_base.fs.dir_exists(self.runner_base.get_job_output_dir(job_id)) == True):
                    utils.warn("ReprocessBatchWorker: resubmit_specific_jobs: deleting files in old directory: job_id: {}, use_checkpoint: {}".format(job_id, use_checkpoint))

                    # delete files
                    for file in self.runner_base.fs.list_files(self.runner_base.get_job_output_dir(job_id)):
                        utils.warn("ReprocessBatchWorker: resubmit_specific_jobs: deleting file: {}".format(file))
                        self.runner_base.fs.delete_file("{}/{}".format(self.runner_base.get_job_output_dir(job_id), file))

            # debug
            utils.info("ReprocessBatchWorker: resubmit_specific_jobs: Job Submitted: {}".format(job_id))

# Class Worker
class Worker:
    def __init__(self, base_dir, worker_id, run_job_func = None, retry_attempts = DEFAULT_RETRY_ATTEMPTS, retry_wait_seconds = DEFAULT_RETRY_WAIT_SEC,
        wait_sec = DEFAULT_WAIT_30SEC, num_iter = 1000):
        self.runner_base = RunnerBase(base_dir)
        self.worker_id = worker_id
        self.run_job_func = run_job_func
        self.retry_attempts = retry_attempts
        self.retry_wait_seconds = retry_wait_seconds
        self.wait_sec = wait_sec
        self.num_iter = num_iter

        # debug
        utils.info("Worker: base_dir: {}, worker_id: {}, retry_attempts: {}, retry_wait_seconds: {}, wait_sec: {}, num_iter: {}".format(base_dir,
            retry_attempts, retry_wait_seconds, worker_id, wait_sec, num_iter))

    def run_loop(self):
        # check worker exists
        if (self.runner_base.fs.dir_exists(self.runner_base.get_worker_dir(self.worker_id)) == False):
            utils.error("Worker is not initialized: {}: {}".format(self.worker_id, self.runner_base.get_worker_dir(self.worker_id)))
            return

        # run iterations
        for i in range(self.num_iter):
            # check for shutdown mode
            if (self.runner_base.is_shutdown_mode() == True):
                utils.warn("Worker: shutdown mode detected. Stopping")
                return

            # run
            utils.run_noreturn_func_with_retry(self.retry_attempts, self.retry_wait_seconds, self.run)

            # sleep
            utils.info("Worker: iteration: {}, Sleeping for {} seconds".format(i, self.wait_sec))
            time.sleep(self.wait_sec)

    def run(self):
        # get job ids
        job_ids = self.runner_base.fs.list_leaf_dir(self.runner_base.get_worker_jobs_dir(self.worker_id))
        utils.info("Worker: worker_id: {}, job_ids: {}".format(self.worker_id, job_ids))

        # run the job
        for job_id in job_ids:
            # read job spec
            created_job_spec_file = self.runner_base.get_job_status_created_job_id_file(job_id)
            running_job_spec_file = self.runner_base.get_job_status_running_job_id_file(job_id)
            completed_job_spec_file = self.runner_base.get_job_status_completed_job_id_file(job_id)
            job_spec = self.runner_base.fs.read_file_contents_as_text(running_job_spec_file)

            # empty file
            if (len(job_spec) == 0):
                job_spec = self.runner_base.fs.read_file_contents_as_text(created_job_spec_file)

            # read json
            job_spec_json = json.loads(job_spec)

            # create the output directory first
            self.runner_base.fs.create_dir(self.runner_base.get_job_output_dir(job_id))

            # run the job
            if (self.run_job_func is not None):
                status = self.run_job_func(job_spec_json)
            else:
                utils.warn("Worker: job_id: {}, run_job_func is None".format(job_id))

            # update the status
            job_spec_json["status"] = status
            job_spec_json["completion_ts"] = timefuncs.utctimestamp_to_datetime_str(timefuncs.get_utctimestamp_sec())

            # write to completed
            self.runner_base.fs.write_text_file(completed_job_spec_file, json.dumps(job_spec_json))

            # debug
            utils.info("Worker: {}, Job Finished: {}, details: {}".format(self.worker_id, job_id, job_spec_json))

            # remove from the local jobs directory
            self.runner_base.fs.delete_file(self.runner_base.get_worker_job_id_file(self.worker_id, job_id))

# class RunAnalysis
class RunAnalysis:
    def __init__(self, base_dir, run_job_func = None, last_n = 200, wait_sec = DEFAULT_WAIT_60SEC):
        self.runner_base = RunnerBase(base_dir)
        self.run_job_func = run_job_func
        self.last_n = last_n
        self.wait_sec = wait_sec

    def run_loop(self):
        # look into assigned jobs directory
        try:
            while True:
                # check for shutdown mode
                if (self.runner_base.is_shutdown_mode() == True):
                    utils.warn("RunAnalysis: : shutdown mode detected. Stopping")
                    return

                # run
                self.run()

                # sleep
                utils.info("RunAnalysis: Sleeping for {} seconds".format(self.wait_sec))
                time.sleep(self.wait_sec)
        except Exception as e:
            utils.error("RunAnalysis: run_loop: caught exception or interrupt: {}".format(e))
            return
        except:
            utils.error("RunAnalysis: run_loop: caught interrupt. Returning")
            return

    def run(self):
        # check the list of running, completed
        all_completed = self.runner_base.fs.list_leaf_dir(self.runner_base.get_job_status_completed())

        # get job ids
        job_ids = sorted(all_completed, reverse = True)[0:self.last_n]

        # iterate
        for job_id in job_ids:
            # check if it exists
            analysis_file = self.runner_base.get_job_graph_analysis_file(job_id)
            if (tsv.exists(analysis_file) == True):
                # check for new columns
                if (RUN_VERSION in tsv.read(analysis_file).get_columns()):
                    utils.debug("RunAnalysis: job_id: {} already exists".format(job_id))
                    continue

            # debug
            utils.info("RunAnalysis: processing job_id: {}: {}".format(job_id, analysis_file))

            # run
            running_job_spec_file = self.runner_base.get_job_status_running_job_id_file(job_id)
            job_spec = self.runner_base.fs.read_file_contents_as_text(running_job_spec_file)

            # failsafe
            if (self.run_job_func is not None):
                self.run_job_func(job_spec)
            else:
                utils.warn("RunAnalysis: run_job_func is None, job_id: {}".format(job_id))





def myfunc_inner_args(x):
    def __myfunc_inner__(y):
        return "Hello World: {}, {}".format(x, y)

    return __myfunc_inner__

def myfunc_inner_kwargs(x = 0):
    def __myfunc_inner__(y):
        return "Hello World: {}, {}".format(x, y)

    return __myfunc_inner__

def myfunc_inner_args_kwargs(x, z = 0):
    def __myfunc_inner__(y):
        return "Hello World: {}, {}, {}".format(x, y, z)

    return __myfunc_inner__

def myfunc_inner_noargs():
    def __myfunc_inner__(y):
        return "Hello World: {}".format(y)

    return __myfunc_inner__

def myfunc_noargs():
    return "Hello World"

# Enforcing module/class/func structure
class ClusterBuiltinFuncs:
    def builtin_sum(arr):
        return sum(arr)

    def builtin_abs(x):
        return abs(x)

    def builtin_len(x):
        return len(x)

    def builtin_min(x, y):
        return min(x, y)

    def builtin_max(x, y):
        return max(x, y)

    def builtin_set(x, y):
        return set(x, y)

    def builtin_ceil(x):
        return math.ceil(x)

    def builtin_floor(x):
        return math.floor(x)

def __step6_explode_func__(mp):
    results = []
    results.append({"label": "unresolved", "count": str(int(mp["total_hosts"]) - int(mp["resolved_hosts"]))})
    results.append({"label": "resolved", "count": mp["resolved_hosts"]})
    return results




# These are the integration parameters with omigo_arjun

# Meta parameters
OMIGO_ARJUN_START_TS = "omigo.arjun.start_ts"
OMIGO_ARJUN_END_TS = "omigo.arjun.end_ts"
OMIGO_ARJUN_EVENT_TS = "omigo.arjun.event_ts"
OMIGO_ARJUN_BASE_PATH = "omigo.arjun.base_path"

OMIGO_ARJUN_INPUT_FILE = "omigo.arjun.input_file"
OMIGO_ARJUN_OUTPUT_FILE = "omigo.arjun.output_file"

# ETL specific header parameters
OMIGO_ARJUN_ETL_PATH_PREFIX = ".omigo.arjun.etl.path."
OMIGO_ARJUN_ETL_START_TS = ".omigo.arjun.etl.start_ts"
OMIGO_ARJUN_ETL_END_TS = ".omigo.arjun.etl.end_ts"
# OMIGO_ARJUN_ETL_USE_FULL_DATA = ".omigo.arjun.etl.use_full_data" # TODO: move this to context
OMIGO_ARJUN_ETL_FILE_PREFIX = "output"

# meta parameters as template
OMIGO_ARJUN_START_TS_TEMPLATE = "{" + OMIGO_ARJUN_START_TS + "}"
OMIGO_ARJUN_END_TS_TEMPLATE = "{" + OMIGO_ARJUN_END_TS + "}"
OMIGO_ARJUN_BASE_PATH_TEMPLATE = "{" + OMIGO_ARJUN_BASE_PATH + "}"

# TODO: these templates dont support multiple inputs
OMIGO_ARJUN_INPUT_FILE_TEMPLATE = "{" + OMIGO_ARJUN_INPUT_FILE + "}"
OMIGO_ARJUN_OUTPUT_FILE_TEMPLATE = "{" + OMIGO_ARJUN_OUTPUT_FILE + "}"

# get etl path by id
def get_etl_path_by_id(strid):
    return "{}{}".format(OMIGO_ARJUN_ETL_PATH_PREFIX, strid)

# is etl path
def is_etl_path(path):
    # check for none
    if (path is None):
        return False

    # return
    return path.startswith(OMIGO_ARJUN_ETL_PATH_PREFIX)

# parse id from etl path
def parse_id_from_etl_path(path):
    # check for none
    if (path is None):
        return None

    # get index
    index = path.find(OMIGO_ARJUN_ETL_PATH_PREFIX)
    if (index != -1):
        return path[index+1:]
    else:
        return None
